/*
 * formel.hpp
 * wtab version 0.02
 * 
 * Copyright 2019 S. Krause <sebkrause@sebkrause.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 */

#pragma once
#include "junktor.hpp"
#include <string>
#include <set>
#include <vector>
#include <exception>
#include <map>

/**
 * \brief Klasse, die die Formeln verwaltet
 * 
 * Diese Klasse erzeugt aus Strings aussagenlogische Formeln und
 * berechnet Werte und erstellt Wahrheitswerttabellen
 */
class Formel
{
public:
    /**
     * \brief Konstruktor - erzeugt Formelobjekt(e)
     * 
     * Der Konstruktor erzeugt aus einem String (der Formel) ein Objekt,
     * indem er es rekursiv immer weiter zerlegt: Aus der Formel
     * "a && !b)" erzeugt er ein Objekt, das genau einen Operator
     * enthält (den mit der niedrigsten Bindung, hier &&), sowie zwei
     * neue Formelobjekte, hier aus "a" und "!b". Ersteres wird als
     * Atom erkannt (das hier als "Atom-Operator" realisiert ist), 
     * letzteres wird als Operator ! und eine weitere Formel "b" zer-
     * legt (das dann ebenfalls ein Atom ist).
     * 
     * Zu den einzelnen Operatoren siehe junktor.hpp, dort sind sie
     * definiert (und die Liste lässt sich einfach erweitern).
     * 
     * @param string Die Formel, wie sie vom User eingegeben wurde
     * @see junktor.hpp
     */
    Formel(std::string);
    
    /**
     * \brief Entfernt Leerzeichen und äußere Klammern
     * 
     * Die Funktion entfernt sämtliche Leerzeichen aus dem übergebenen
     * String. Danach entfernt sie Klammerpaare, die die komplette
     * Formel umschließen. Aus '((a=>!(a&&b)))' wird also 'a=>!(a&&b)',
     * wohingegen '(!a)=>(b)' nicht verändert wird.
     * Die syntaktische Korrektheit der Klammern wird nicht überprüft,
     * es wird stur '((((a)' zu '(((a'.
     * 
     * @param string Die übergebene Zeichenkette
     * @return Die bereinigte Zeichenkette
     */
    static std::string trim(std::string);
    
    /**
     * \brief Checkt, ob ein String eine Variable sein könnte
     * 
     * Die Funktion checkt, ob ein String eine Variable ist: Dazu muss er
     * genau einen Buchstaben und danach beliebig Ziffern enthalten,
     * sonst nichts.
     * 
     * @param s Der zu prüfende String
     * @return true, wenn c ein Buchstabe, false sonst
     */
    static bool istVariable(const std::string& s);
    
    
    /* \brief Ist das Zeichen c ein Buchstabe?
     * 
     * Die Funktion checkt, ob das Zeichen im Bereich von
     * a-z oder A-Z liegt.
     *
     * @param c Das zu prüfende Zeichen
     * @return true, falls c Buchstabe, false sonst
     */
    static bool istBuchstabe(char c);
    
    /* \brief Ist das Zeichen c eine Ziffer?
     * 
     * Die Funktion checkt, ob das Zeichen im Bereich von
     * 0-9 liegt.
     *
     * @param c Das zu prüfende Zeichen
     * @return true, falls c Ziffer, false sonst
     */
    static bool istZiffer(char c);
    
    /**
     * \brief Der Wert der Formel unter einer Belegung
     * 
     * Die Funktion berechnet den Wert der Formel unter einer als Map
     * gegebenen Belegung, indem sie entweder den Wert ihrer Variablen
     * aus der Belegung liest und zurückgibt (falls sie ein Atom ist)
     * oder den/die Wert(e) ihrer Kinder mithilfe ihres Operators bzw.
     * Junktors verknüpft und dieses Ergebnis zurückgibt. So entsteht
     * rekursiv der Gesamtwert einer komplexeren Formel.
     * 
     * Werte werden hier generell als Integer und nicht als Bool be-
     * handelt - interessant evtl. für Fuzzy-Logiken o.ä.
     * 
     * @param belegung Eine Map mit allen Variablen und deren Werten
     * @return Der Wert der Formel unter der Belegung
     */
    int wert(std::map<std::string, int> belegung) const;
    
    /**
     * \brief Gibt eine Wahrheitswerttabelle auf der Konsole aus
     * 
     * Die Funktion berechnet den Wert der Formel für jede mögliche
     * Belegung und gibt diese in einer Wertetabelle aus. Anschließend
     * werden noch die Modellmenge und die Information, ob es sich um
     * eine Kontradiktion oder eine Tautologie handelt, ausgegeben.
     * 
     * Die Werte werden mit Formel::wert() berechnet.
     * */
    void werteTabelle() const;
    
    /* Die Formel, wie vom User eingegeben */
    std::string originalstring;
    
private:
    /* Der Operator der Formel (es gibt pro Formel genau einen) */
    junkt * junktor;
    
    /* Ist die Formel ein Atom? */
    bool istatom;
    
    /* Die Menge an Variablen, die in der Formel vorkommen */
    std::set<std::string> vars;
    
    /* Ein Vektor mit 0, 1 oder 2 Kindern der Formel */
    std::vector<Formel*> kindformeln;
};
